#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <cairo.h>
#include <glib.h>

#include "peswrite.h"


void
pes_cmd_move_to (PESContext *context,
                 gdouble x, gdouble y)
{
  PESCommand *cmd = g_new0 (PESCommand, 1);

  cmd->x = context->cur_x = x;
  cmd->y = context->cur_y = y;
  cmd->cmd = CMD_MOVE_TO;
  cmd->x_stitch = rint (cmd->x * context->scale + context->x0);
  cmd->y_stitch = rint (cmd->y * context->scale + context->x0);

  context->commands = g_list_append (context->commands, cmd);
}


void
pes_cmd_line_to (PESContext *context,
                 gdouble x, gdouble y)
{
  PESCommand *cmd = g_new0 (PESCommand, 1);

  cmd->x = context->cur_x = x;
  cmd->y = context->cur_y = y;
  cmd->cmd = CMD_LINE_TO;
  cmd->x_stitch = rint (cmd->x * context->scale + context->x0);
  cmd->y_stitch = rint (cmd->y * context->scale + context->x0);

  context->commands = g_list_append (context->commands, cmd);
}


void
pes_cmd_rel_move_to (PESContext *context,
                     gdouble x, gdouble y)
{
  PESCommand *cmd = g_new0 (PESCommand, 1);

  cmd->x = context->cur_x += x;
  cmd->y = context->cur_y += y;
  cmd->cmd = CMD_MOVE_TO;
  cmd->x_stitch = rint (cmd->x * context->scale + context->x0);
  cmd->y_stitch = rint (cmd->y * context->scale + context->x0);

  context->commands = g_list_append (context->commands, cmd);
}


void
pes_cmd_rel_line_to (PESContext *context,
                     gdouble x, gdouble y)
{
  PESCommand *cmd = g_new0 (PESCommand, 1);

  cmd->x = context->cur_x += x;
  cmd->y = context->cur_y += y;
  cmd->cmd = CMD_LINE_TO;
  cmd->x_stitch = rint (cmd->x * context->scale + context->x0);
  cmd->y_stitch = rint (cmd->y * context->scale + context->x0);

  context->commands = g_list_append (context->commands, cmd);
}



static int
pes_encode_numbers (guint8 *cmdbuf, gint x, gint y)
{
  gint pos = 0;

  if (x < -2048 || x > 2047 || y < -2048 || y > 2047)
    g_printerr ("coords out of range: x=%d, y=%d\n", x, y);

  if (x >= 0)
    {
      if (x >= 64)
        { 
          cmdbuf[pos++] = ((x >> 8) & 0x07) | 0x80;
          cmdbuf[pos++] = x & 0xff;
        }
      else
        {
          cmdbuf[pos++] = x & 0x3f;
        }
    }
  else
    {
      if (x < -64)
        {
          x += 4096;
          cmdbuf[pos++] = ((x >> 8) & 0x07) | 0x88;
          cmdbuf[pos++] = x & 0xff;
        }
      else
        {
          x += 128;
          cmdbuf[pos++] = (x & 0x3f) | 0x40;
        }
    }

  if (y >= 0)
    {
      if (y >= 64)
        { 
          cmdbuf[pos++] = ((y >> 8) & 0x07) | 0x80;
          cmdbuf[pos++] = y & 0xff;
        }
      else
        {
          cmdbuf[pos++] = y & 0x3f;
        }
    }
  else
    {
      if (y < -64)
        {
          y += 4096;
          cmdbuf[pos++] = ((y >> 8) & 0x07) | 0x88;
          cmdbuf[pos++] = y & 0xff;
        }
      else
        {
          y += 128;
          cmdbuf[pos++] = (y & 0x3f) | 0x40;
        }
    }

  return pos;
}

void
pes_export_pec (PESContext *context,
                FILE          *outfile)
{
  GList *list;
  PESCommand *cmd;
  guint8  cmdbuf[4] = { 0, }, pechdr[528] = { 0, };
  glong pecstart, graphicstart;
  gint last_x, last_y, width, height;
  gint ret;

  pecstart = ftell (outfile);

  memset (pechdr, 0x20, sizeof (pechdr));
  memcpy (pechdr, "LA:12345678", 11);

  pechdr[19] = 0x0d;
  pechdr[31] = 0xff;
  pechdr[32] = 0xff;
  pechdr[33] = 0x00;
  pechdr[34] = 0x06;
  pechdr[35] = 0x26;
  pechdr[40] = 0x64;
  pechdr[42] = 0x00;
  pechdr[44] = 0x00;
  pechdr[45] = 0xff;
  pechdr[48] = 0;  // 1 colors
  pechdr[49] = 20;   // black

  pechdr[512] = 0x00;
  pechdr[513] = 0x00;

  /* preview graphics start location (relative to pec+512).
   * 3 bytes, will be filled in later */
  pechdr[514] = 0x00;
  pechdr[515] = 0x00;
  pechdr[516] = 0x00;

  pechdr[517] = 0x31;
  pechdr[518] = 0xff;
  pechdr[519] = 0xf0;

  width = context->max_x - context->min_x;
  height = context->max_y - context->min_y;
  pechdr[520] = width & 0xff;
  pechdr[521] = (width >> 8) & 0xff;
  pechdr[522] = height & 0xff;
  pechdr[523] = (height >> 8) & 0xff;
  pechdr[524] = 0xe0;
  pechdr[525] = 0x01;
  pechdr[526] = 0xb0;
  pechdr[527] = 0x01;

  fwrite (pechdr, sizeof (guint8), sizeof (pechdr), outfile);

  list = g_list_first (context->commands);
  cmd = list->data;
  last_x = cmd->x_stitch - context->min_x;
  last_y = cmd->y_stitch - context->min_y;

  /* movement to first point */
  cmdbuf[0] = ((last_x >> 8) & 0x0f) | 0x90;
  cmdbuf[1] = last_x & 0xff;
  cmdbuf[2] = ((last_y >> 8) & 0x0f) | 0x90;
  cmdbuf[3] = last_y & 0xff;
  fwrite (cmdbuf, sizeof (guint8), 4, outfile);

  last_x = cmd->x_stitch;
  last_y = cmd->y_stitch;

  for (list = g_list_next (list);
       list;
       list = g_list_next (list))
    {
      cmd = list->data;

      switch (cmd->cmd)
        {
          case CMD_MOVE_TO:
            ret = pes_encode_numbers (cmdbuf,
                                      cmd->x_stitch - last_x,
                                      cmd->y_stitch - last_y);
            fwrite (cmdbuf, sizeof (guint8), ret, outfile);
            last_x = cmd->x_stitch;
            last_y = cmd->y_stitch;
            break;
            
          case CMD_LINE_TO:
            ret = pes_encode_numbers (cmdbuf,
                                      cmd->x_stitch - last_x,
                                      cmd->y_stitch - last_y);
            fwrite (cmdbuf, sizeof (guint8), ret, outfile);
            last_x = cmd->x_stitch;
            last_y = cmd->y_stitch;
            break;

          default:
            g_printerr ("Huh?\n");
            break;
        }
    }

  fwrite ("\xff", sizeof (guint8), 1, outfile);

  graphicstart = ftell (outfile);

  fseek (outfile, pecstart + 514, SEEK_SET);
  cmdbuf[0] = (graphicstart - pecstart - 512) & 0xff;
  cmdbuf[1] = ((graphicstart - pecstart - 512) >> 8) & 0xff;
  cmdbuf[2] = ((graphicstart - pecstart - 512) >> 16) & 0xff;
  fwrite (cmdbuf, sizeof (guint8), 3, outfile);

  fseek (outfile, 0, SEEK_END);
}

static void
pes_write_uint16 (FILE *outfile,
                  int   i)
{
  char data[2];

  data[0] = i & 0xff;
  data[1] = (i >> 8) & 0xff;

  fwrite (data, sizeof (char), 2, outfile);
}


static void
pes_write_float (FILE *outfile,
                 float f)
{
  fwrite (&f, sizeof (float), 1, outfile);
}


static void
pes_write_string (FILE *outfile,
                  char *string,
                  int   len)
{
  char l[2];
  if (len < 0)
    len = strlen (string);

  l[0] = MIN (255, len);
  fwrite (l, sizeof (char), 1, outfile);

  if (len > 254)
    {
      l[0] = len & 0xff;
      l[1] = (len >> 8) & 0xff;
      fwrite (l, sizeof (char), 2, outfile);
    }
  fwrite (string, sizeof (char), len, outfile);
}


static void
pes_write_colortable (FILE *outfile, int num_colors)
{
  int i;

  pes_write_uint16 (outfile, 1);

  /* "You can have it any color you like, as long as it's black." */
  for (i = 0; i < num_colors; i++)
    {
      /* color */
      pes_write_string (outfile, "900", -1);
      /* RGB representation */
      fwrite ("\x00\x00\x00", sizeof (char), 3, outfile);
      /* unknown data */
      fwrite ("\x00\x0a\x00\x00\x00", sizeof (char), 5, outfile);
      pes_write_string (outfile, "20", -1);
      pes_write_string (outfile, "EMBROIDERY", -1);
      pes_write_string (outfile, "", -1);
    }
}


static void
pes_write_block_header (FILE *outfile, char *block_name)
{
  int l;
  /* unknown pre-header data */
  pes_write_uint16 (outfile, 1);
  fwrite ("\xff\xff\x00\x00", sizeof (char), 4, outfile);
  l = strlen (block_name);
  pes_write_uint16 (outfile, l);
  fwrite (block_name, sizeof (char), l, outfile);
}


static void
pes_write_block_CEmbOne (FILE *outfile,
                         PESContext *context)
{
  pes_write_block_header (outfile, "CEmbOne");

  pes_write_uint16 (outfile, context->min_x);
  pes_write_uint16 (outfile, context->min_y);
  pes_write_uint16 (outfile, context->max_x);
  pes_write_uint16 (outfile, context->max_y);
  pes_write_uint16 (outfile, context->min_x);
  pes_write_uint16 (outfile, context->min_y);
  pes_write_uint16 (outfile, context->max_x);
  pes_write_uint16 (outfile, context->max_y);

  /* transformation matrix */
  pes_write_float (outfile, 1.0);
  pes_write_float (outfile, 0.0);
  pes_write_float (outfile, 0.0);
  pes_write_float (outfile, 1.0);
  pes_write_float (outfile, 0.0);
  pes_write_float (outfile, 0.0);

  pes_write_uint16 (outfile, 1);
  pes_write_uint16 (outfile, context->min_x);
  pes_write_uint16 (outfile, context->max_y);
  pes_write_uint16 (outfile, context->max_x - context->min_x);
  pes_write_uint16 (outfile, context->max_y - context->min_y);
  pes_write_uint16 (outfile, 0);
  pes_write_uint16 (outfile, 0);
  pes_write_uint16 (outfile, 0);
  pes_write_uint16 (outfile, 0);
}


static void
pes_write_block_CSewSeg (FILE       *outfile,
                         PESContext *context)
{
  GList *list;
  PESCommand *cmd;

  pes_write_block_header (outfile, "CSewSeg");

  pes_write_uint16 (outfile, 0);
  pes_write_uint16 (outfile, 0);

  pes_write_uint16 (outfile, g_list_length (context->commands));

  for (list = g_list_first (context->commands);
       list;
       list = g_list_next (list))
    {
      cmd = list->data;

      switch (cmd->cmd)
        {
          case CMD_MOVE_TO:
          case CMD_LINE_TO:
            pes_write_uint16 (outfile, cmd->x_stitch);
            pes_write_uint16 (outfile, cmd->y_stitch);
            break;
          default:
            g_printerr ("Huh?\n");
            break;
        }
    }

  fwrite ("\x01\x00"
          "\x00\x00"
          "\x00\x00"
          "\x00\x00"
          "\x00\x00"
          "\x00\x00"
          "\x00\x00"
          "\x00\x00\x00\x00\x00\x00\x00\x00",
          sizeof (char), 22, outfile);
}


void
pes_draw_render (PESContext *context,
                 cairo_t    *cr,
                 gdouble     width,
                 gdouble     height,
                 gboolean    show_starts)
{
  GList *list;
  PESCommand *cmd;
  gdouble scaling, x0, y0, w, pixelwidth;
  gdouble size_x, size_y;

  size_x = context->max_x - context->min_x;
  size_y = context->max_y - context->min_y;

  cairo_translate (cr, width / 2, height / 2);
  scaling = MIN (width, height) / 2.2;
  cairo_scale (cr, scaling, -scaling);
  cairo_translate (cr, -1, 1);
  w = MIN (2.0 / size_x, 2.0 / size_y);
  cairo_scale (cr, w, -w);
  cairo_translate (cr, -context->x0, -context->y0);

  x0 = y0 = 1.0;
  cairo_device_to_user_distance (cr, &x0, &y0);
  pixelwidth = MAX (x0, y0);

  cairo_set_font_size (cr, pixelwidth * 14);

  cairo_set_source_rgba (cr, 1.0, 1.0, 1.0, 1.0);
  cairo_set_line_width (cr, pixelwidth);
  cairo_set_line_join (cr, CAIRO_LINE_JOIN_ROUND);

  for (list = g_list_first (context->commands);
       list;
       list = g_list_next (list))
    {
      cmd = list->data;

      switch (cmd->cmd)
        {
          case CMD_MOVE_TO:
            cairo_stroke (cr);
            if (show_starts)
              {
                cairo_set_source_rgba (cr, 1.0, 0.0, 0.0, 1.0);
                cairo_arc (cr, cmd->x_stitch, cmd->y_stitch, pixelwidth * 5,
                           0, 2*G_PI);
                cairo_stroke (cr);
                cairo_set_source_rgba (cr, 1.0, 1.0, 1.0, 1.0);
              }
            cairo_move_to (cr, cmd->x_stitch, cmd->y_stitch);
            break;

          case CMD_LINE_TO:
            cairo_line_to (cr, cmd->x_stitch, cmd->y_stitch);
            break;

          default:
            g_printerr ("Huh?\n");
            break;
        }
    }

  cairo_stroke (cr);
}


void
pes_export_preview (PESContext *context,
                    FILE       *outfile)
{
  cairo_surface_t *surf;
  cairo_t *cr;
  guchar *line;
  int i;

  surf = cairo_image_surface_create (CAIRO_FORMAT_A1, 48, 38);
  cr = cairo_create (surf);

  cairo_set_line_width (cr, 1.0);
  cairo_move_to (cr, 4.5, 1.5);
  cairo_line_to (cr, 48 - 4.5, 1.5);
  cairo_line_to (cr, 48 - 1.5, 4.5);
  cairo_line_to (cr, 48 - 1.5, 38 - 4.5);
  cairo_line_to (cr, 48 - 4.5, 38 - 1.5);
  cairo_line_to (cr, 4.5, 38 - 1.5);
  cairo_line_to (cr, 1.5, 38 - 4.5);
  cairo_line_to (cr, 1.5, 4.5);
  cairo_close_path (cr);
  cairo_stroke (cr);

  cairo_translate (cr, 3, 3);
  pes_draw_render (context, cr, 42, 32, FALSE);

  cairo_destroy (cr);
  cairo_surface_write_to_png (surf, "preview.png");

  line = cairo_image_surface_get_data (surf);
  for (i = 0; i < cairo_image_surface_get_height (surf); i++)
    {
      fwrite (line, sizeof (char), 6, outfile);
      line += cairo_image_surface_get_stride (surf);
    }
  cairo_surface_destroy (surf);
}


void
pes_export_name_preview (PESContext *context,
                         FILE       *outfile)
{
  cairo_surface_t *surf;
  cairo_t *cr;
  guchar *line;
  int i;

  surf = cairo_image_surface_create (CAIRO_FORMAT_A1, 48, 24);
  cr = cairo_create (surf);

   cairo_select_font_face (cr, "sans-serif",
                           CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
  cairo_set_font_size (cr, 16);
  cairo_move_to (cr, 2.5, 18);
  cairo_show_text (cr, "huhu");

  cairo_destroy (cr);
  cairo_surface_write_to_png (surf, "previewtext.png");

  line = cairo_image_surface_get_data (surf);
  for (i = 0; i < cairo_image_surface_get_height (surf); i++)
    {
      fwrite (line, sizeof (char), 6, outfile);
      line += cairo_image_surface_get_stride (surf);
    }
  cairo_surface_destroy (surf);
}


void
pes_export_pes (PESContext *context,
                FILE       *outfile)
{
  GList *list;
  PESCommand *cmd;
  guint32 pecstart;
  gint i;

  pecstart = 0x0;

  /* header */
  fwrite ("#PES0070", sizeof (char), 8, outfile);
  /* PEC-pointer placeholder */
  fwrite (&pecstart, sizeof (guint32), 1, outfile);  /* placeholder */

  context->min_x = 32768;
  context->max_x = -32768;
  context->min_y = 32768;
  context->max_y = -32768;

  for (list = g_list_first (context->commands);
       list;
       list = g_list_next (list))
    {
      cmd = list->data;
      context->min_x = MIN (context->min_x, cmd->x_stitch);
      context->max_x = MAX (context->max_x, cmd->x_stitch);
      context->min_y = MIN (context->min_y, cmd->y_stitch);
      context->max_y = MAX (context->max_y, cmd->y_stitch);
    }

  /* unknown data */
  fwrite ("\x01\x00\x30\x31", sizeof (char), 4, outfile);

  pes_write_string (outfile, "", -1);
  pes_write_string (outfile, "", -1);
  pes_write_string (outfile, "", -1);
  pes_write_string (outfile, "", -1);
  pes_write_string (outfile, "", -1);

  pes_write_uint16 (outfile, 0);     /* optimize hoop change */
  pes_write_uint16 (outfile, 0);     /* page size == hoop size */
  pes_write_uint16 (outfile, 100);   /* hoop width in mm */
  pes_write_uint16 (outfile, 100);   /* hoop height in mm */
  pes_write_uint16 (outfile, 0);     /* display option */
  pes_write_uint16 (outfile, 100);   /* page width in mm */
  pes_write_uint16 (outfile, 100);   /* oop width in mm */
  pes_write_uint16 (outfile, 100);   /* hoop height in mm */
  pes_write_uint16 (outfile, 100);   /* hoop height in mm */
  pes_write_uint16 (outfile, 100);   /* unknown */
  pes_write_uint16 (outfile, 7);     /* background color index */
  pes_write_uint16 (outfile, 19);    /* page color index */
  pes_write_uint16 (outfile, 0);     /* show grid */
  pes_write_uint16 (outfile, 0);     /* grid shown with axes */
  pes_write_uint16 (outfile, 0);     /* snap to grid */
  pes_write_uint16 (outfile, 25);    /* grid interval (* 0.1mm) */
  pes_write_uint16 (outfile, 1);     /* Sewing area: Design Page area */
  pes_write_uint16 (outfile, 0);     /* optimize entry points */

  /* unknown data */
  fwrite ("\x00"
          "\x00\x00\x80\x3f"
          "\x00\x00\x00\x00"
          "\x00\x00\x00\x00"
          "\x00\x00\x80\x3f"
          "\x00\x00\x00\x00"
          "\x00\x00\x00\x00"
          "\x00\x00\x00\x00"
          "\x00\x00"
          , sizeof (char), 31, outfile);

  /* ok, ignoring a whole bunch of unknown data here */

  /* color table */

  pes_write_colortable (outfile, 1);

  pes_write_block_CEmbOne (outfile, context);
  pes_write_block_CSewSeg (outfile, context);

  /* fill in PEC pointer placeholder */
  pecstart = ftell (outfile);
  fseek (outfile, 0x08, SEEK_SET);
  fwrite (&pecstart, sizeof (guint32), 1, outfile);
  fseek (outfile, 0x0, SEEK_END);

  pes_export_pec (context, outfile);
  pes_export_preview (context, outfile);
  pes_export_preview (context, outfile);

  fwrite ("\x00\x14\x20\x20\x20\x20\x20\x20", sizeof (char), 8, outfile);
  for (i = 0; i < 15; i++)
    fwrite ("\x20\x20\x20\x20\x20\x20\x20\x20", sizeof (char), 8, outfile);

  pes_export_name_preview (context, outfile);

  fwrite ("\x00\x00\x00\x00\x00\x0D\x84\x03", sizeof (char), 8, outfile);

}


