#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <cairo.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "peswrite.h"

#define PIXEL_VISIBLE  (1 << 0)
#define PIXEL_VISITED  (1 << 1)

#define STEP_NUM (7)
#define STEP_WIDTH (0.5 / (2 * STEP_NUM))

typedef struct _pstick_context PStickContext;

struct _pstick_context
{
  gdouble       pixelwidth;

  gboolean      fullscreen;

  GtkWidget    *window;
  GtkWidget    *darea;
  cairo_t      *cr;

  gint          size_x;
  gint          size_y;
  gint         *pixels;
  gint          in_stroke;

  PESContext   *pes;
};



void
pstick_draw_begin (PStickContext *context)
{
  g_return_if_fail (context->cr == NULL);

  context->cr = gdk_cairo_create (context->darea->window);

  cairo_set_source_rgba (context->cr, 0.7, 0.7, 0.7, 1.0);
  cairo_paint (context->cr);
}


void
pstick_draw_end (PStickContext *context)
{
  cairo_destroy (context->cr);
  context->cr = NULL;
}


void
pstick_dfs_next_step (PStickContext *context,
                      gint x, gint y, gint enterpt)
{
  gint i, j;

  if (x < 0 || y < 0 || x >= context->size_x || y >= context->size_y ||
      !(context->pixels [y * context->size_x + x] & PIXEL_VISIBLE) ||
       (context->pixels [y * context->size_x + x] & PIXEL_VISITED))
    return;

  if (enterpt < 0)
    {
      if (!context->in_stroke)
        pes_cmd_move_to (context->pes, x, y);

      pes_cmd_line_to (context->pes, x, y);
      enterpt = 0;
      context->in_stroke = TRUE;
    }

  context->pixels [y * context->size_x + x] |= PIXEL_VISITED;

  for (i = 0; i < 8; i++)
    {
      switch (enterpt)
        {
          case 0:
            for (j = 0; j < 2*STEP_NUM; j++)
              pes_cmd_rel_line_to (context->pes, STEP_WIDTH, (j % 2) ? -0.5 : 0.5);
            enterpt = 1;
            pstick_dfs_next_step (context, x, y - 1, 7 - enterpt);
            break;
          case 1:
            pes_cmd_rel_line_to (context->pes, 0.5, 0.0);
            enterpt = 2;
            pstick_dfs_next_step (context, x + 1, y - 1, 7 - enterpt);
            break;
          case 2:
            for (j = 0; j < 2*STEP_NUM; j++)
              pes_cmd_rel_line_to (context->pes, j % 2 ? 0.5 : -0.5, STEP_WIDTH);
            enterpt = 3;
            pstick_dfs_next_step (context, x + 1, y, 7 - enterpt);
            break;
          case 3:
            pes_cmd_rel_line_to (context->pes, 0.0, 0.5);
            enterpt = 7;
            pstick_dfs_next_step (context, x + 1, y + 1, 7 - enterpt);
            break;
          case 7:
            for (j = 0; j < 2*STEP_NUM; j++)
              pes_cmd_rel_line_to (context->pes, -STEP_WIDTH, j % 2 ? 0.5 : -0.5);
            enterpt = 6;
            pstick_dfs_next_step (context, x, y + 1, 7 - enterpt);
            break;
          case 6:
            pes_cmd_rel_line_to (context->pes, -0.5, 0.0);
            enterpt = 5;
            pstick_dfs_next_step (context, x - 1, y + 1, 7 - enterpt);
            break;
          case 5:
            for (j = 0; j < 2*STEP_NUM; j++)
              pes_cmd_rel_line_to (context->pes, j % 2 ? -0.5 : 0.5, -STEP_WIDTH);
            enterpt = 4;
            pstick_dfs_next_step (context, x - 1, y, 7 - enterpt);
            break;
          case 4:
            pes_cmd_rel_line_to (context->pes, 0.0, -0.5);
            enterpt = 0;
            pstick_dfs_next_step (context, x - 1, y - 1, 7 - enterpt);
            break;

          default:
            g_printerr ("huh?\n");
            break;
        }
    }
}


void
pstick_run_dfs (PStickContext *context)
{
  gint x, y;

#if 0
  pes_cmd_move_to (context, 575, 525);
  pes_cmd_line_to (context, 675, 525);
  pes_cmd_line_to (context, 775, 525);
  pes_cmd_line_to (context, 675, 925);

  return;
#endif

  for (y = 0; y < context->size_y; y++)
    for (x = 0; x < context->size_x; x++)
      context->pixels [y * context->size_x + x] &= PIXEL_VISIBLE;

  for (y = 0; y < context->size_y; y++)
    {
      for (x = 0; x < context->size_x; x++)
        {
          context->in_stroke = 0;
          pstick_dfs_next_step (context, x, y, -1);
        }
    }
}



gboolean
pstick_expose_event (GtkWidget      *widget,
                     GdkEventExpose *event,
                     gpointer       *data)
{
  PStickContext *context = (PStickContext *) data;
  gint x, y;
  gdouble x0, y0, scaling, w, pixelwidth;
  GtkAllocation *alloc;

  pstick_draw_begin (context);

  cairo_save (context->cr);

  alloc = &context->darea->allocation;
  scaling = MIN (alloc->width, alloc->height) / 2.2;
  cairo_translate (context->cr, alloc->width / 2, alloc->height / 2);
  cairo_scale (context->cr, scaling, -scaling);
  cairo_translate (context->cr, -1, 1);
  w = MIN (2.0 / context->size_x, 2.0 / context->size_y);
  cairo_scale (context->cr, w, -w);

  x0 = y0 = 1.0;
  cairo_device_to_user_distance (context->cr, &x0, &y0);
  pixelwidth = MAX (x0, y0);

  cairo_set_source_rgba (context->cr, 0, 0, 0, 1.0);
  for (y = 0; y < context->size_y; y++)
    {
      for (x = 0; x < context->size_x; x++)
        {
          if (context->pixels [y * context->size_x + x] & PIXEL_VISIBLE)
            {
              cairo_rectangle (context->cr, x, y, 1, 1);
            }
        }
    }
  cairo_fill (context->cr);

  cairo_restore (context->cr);

  pes_draw_render (context->pes, context->cr, alloc->width, alloc->height, TRUE);

  pstick_draw_end (context);

  return FALSE;
}


void
pstick_button_event (GtkWidget      *widget,
                     GdkEventButton *event,
                     gpointer       *data)
{
  PStickContext *context = (PStickContext *) data;
  gdouble x, y;

  x = event->x;
  y = event->y;

  pstick_draw_begin (context);
  cairo_device_to_user (context->cr, &x, &y);
  pstick_draw_end (context);
}


void
pstick_motion_notify (GtkWidget      *widget,
                      GdkEventMotion *event,
                      gpointer       *data)
{
  return;
}


gboolean
pstick_key_press_event (GtkWidget   *widget,
                        GdkEventKey *event,
                        gpointer    *data)
{
  PStickContext *context = (PStickContext *) data;

  switch (event->keyval)
    {
      case GDK_F:
      case GDK_f:
      case GDK_F11:
        if (context->fullscreen)
          gtk_window_unfullscreen (GTK_WINDOW (context->window));
        else
          gtk_window_fullscreen (GTK_WINDOW (context->window));
        context->fullscreen = !context->fullscreen;
        break;

      case GDK_Q:
      case GDK_q:
      case GDK_Escape:
        gtk_main_quit ();
        break;
      default:
        return FALSE;
    }

  return TRUE;
}


void
pstick_create_window (PStickContext *context)
{
  GtkWidget *vbox;

  context->window = g_object_new (GTK_TYPE_WINDOW,
                                  "title", "PStick",
                                  "default-width", 800,
                                  "default-height", 600,
                                  NULL);

  g_signal_connect (context->window,
                    "destroy", G_CALLBACK (gtk_main_quit), NULL);

  vbox = g_object_new (GTK_TYPE_VBOX,
                       "visible", TRUE,
                       "parent", context->window,
                       NULL);

  context->darea = g_object_new (GTK_TYPE_DRAWING_AREA,
                                 "width-request", 100,
                                 "height-request", 100,
                                 "visible", TRUE,
                                 "parent", vbox,
                                 NULL);
  g_object_connect (G_OBJECT (context->darea),
                    "signal::expose-event",
                        G_CALLBACK (pstick_expose_event), context,
                    "signal::motion_notify_event",
                        G_CALLBACK (pstick_motion_notify), context,
                    "signal::button_press_event",
                        G_CALLBACK (pstick_button_event), context,
                    NULL);

  g_object_connect (G_OBJECT (context->window),
                    "signal::key_press_event",
                        G_CALLBACK (pstick_key_press_event), context,
                    NULL);

  gtk_widget_set_events (context->darea,
                         gtk_widget_get_events (context->darea)
                         | GDK_LEAVE_NOTIFY_MASK
                         | GDK_BUTTON_PRESS_MASK
                         | GDK_BUTTON_RELEASE_MASK
                         | GDK_KEY_PRESS_MASK
                         | GDK_POINTER_MOTION_MASK
                         | GDK_POINTER_MOTION_HINT_MASK);
}


gint
pstick_load_pixels (PStickContext *context,
                    char          *filename)
{
  GdkPixbuf *pixbuf;
  gint x, y, bpp;
  guchar *line, *p;

  pixbuf = gdk_pixbuf_new_from_file (filename, NULL);

  if (!pixbuf)
    {
      g_printerr ("cannot open input image\n");
      exit (1);
    }

  context->size_x = gdk_pixbuf_get_width (pixbuf);
  context->size_y = gdk_pixbuf_get_height (pixbuf);
  context->pixels = g_new0 (gint, context->size_x * context->size_y);

  line = gdk_pixbuf_get_pixels (pixbuf);
  bpp = gdk_pixbuf_get_n_channels (pixbuf);

  for (y = 0; y < context->size_y; y++)
    {
      p = line;
      for (x = 0; x < context->size_x; x++)
        {
          context->pixels [y * context->size_x + x] = *p ? 0 : PIXEL_VISIBLE;
          p += bpp;
        }
      line += gdk_pixbuf_get_rowstride (pixbuf);
    }

  return 1;
}


gint
main (gint argc, char *argv[])
{
  PStickContext *context;

  if (argc < 2 || argc > 3)
    {
      g_printerr ("usage: %s <infile> [<outfile>]\n", argv[0]);
      exit (1);
    }

  gtk_init (&argc, &argv);

  context = g_new0 (PStickContext, 1);
  context->pes = g_new0 (PESContext, 1);
  context->pes->scale = 70;
  context->pes->x0    = 600;
  context->pes->y0    = 600;


  pstick_load_pixels (context, argv[1]);

  pstick_run_dfs (context);

  if (argc == 3)
    {
      FILE *outfile;
      outfile = fopen (argv[2], "w");
      if (!outfile)
        return -1;

      pes_export_pes (context->pes, outfile);
    }

  pstick_create_window (context);
  gtk_widget_show (context->window);

  gtk_main ();

  return 0;
}
